////////////////////////////////////////////////////////////////////////////////
//                                                                             /
// 2012-2020 (c) Baical                                                        /
//                                                                             /
// This library is free software; you can redistribute it and/or               /
// modify it under the terms of the GNU Lesser General Public                  /
// License as published by the Free Software Foundation; either                /
// version 3.0 of the License, or (at your option) any later version.          /
//                                                                             /
// This library is distributed in the hope that it will be useful,             /
// but WITHOUT ANY WARRANTY; without even the implied warranty of              /
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           /
// Lesser General Public License for more details.                             /
//                                                                             /
// You should have received a copy of the GNU Lesser General Public            /
// License along with this library.                                            /
//                                                                             /
////////////////////////////////////////////////////////////////////////////////
#ifndef UP7_PROXY_STREAM_H
#define UP7_PROXY_STREAM_H

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
struct stProxyPacket
{
    stuP7baseHdr  *pHdr;
    stProxyPacket *pNext;
};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class CTimeList: public CListPool<int64_t>
{
public:
    CTimeList() : CListPool(16) {}
protected:
    virtual tBOOL Data_Release(int64_t i_pData) { UNUSED_ARG(i_pData); return TRUE; }
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class IProxyStream
{
public:
    virtual void     EnableTimeSync() = 0;
    virtual void     DisableTimeSync() = 0;
    virtual void     UpdateStartTime() = 0;
    virtual bool     ParseDescription(uint32_t i_uSessionId, const uint8_t *i_pDesc, size_t i_szDesc)  = 0;
    virtual size_t   Process(const stProxyPacket *i_pPackets)  = 0;
    virtual uint64_t GetSupportedPacketsMap() = 0;
    virtual bool     Maintain() = 0;
    virtual int32_t  Add_Ref() = 0;
    virtual int32_t  Release() = 0;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class CProxyStream
    : public virtual IProxyStream
    , public virtual IP7C_Channel    
    , public virtual CChunks
{            
protected:
    enum eState
    {
        eStateDescription,
        eStateTimeSync,
        eStateReady,
        eStateError
    };

    //put volatile variables at the top, to obtain 32 bit alignment. 
    //Project has 8 bytes alignment by default
    tINT32 volatile        m_lReference;
                          
    CWString               m_cName;
    bool                   m_bConvertEndianess;
    IP7_Trace             *m_pP7Trace;
    IP7_Trace::hModule     m_hP7Mod;
    IP7_Telemetry         *m_pP7Tel;
    tUINT16                m_sTelTimeCorrection;
    CuP7Fifo              *m_pFifo;
    uint8_t                m_bId;
    eState                 m_eState;
    uint32_t               m_uSessionId;
                          
    IProxyClient          *m_pProxyClient;
    IP7_Client            *m_pClient;
    uint32_t               m_uClientId;
    sP7C_Status            m_stStatus;
                          
    CLinearPool            m_cPool;

    tLOCK                  m_hLock;

    bool                   m_bClosed;

    uint64_t               m_qwCpuStartTime;
    uint64_t               m_qwCpuFreq;
    int64_t                m_qwCpuHostTimeOffset;

    uint64_t               m_qwHostStartTime;
    uint64_t               m_qwHostFreq;

    CTimeList              m_cCorrectionList;
    int64_t                m_llTimeCorrection;

    uint32_t               m_uOperationTimeStamp;
public:
    CProxyStream(CWString      &i_rName, 
                 bool           i_bConvertEndianess,
                 CuP7Fifo      *i_pFifo,
                 uint8_t        i_bId,
                 uint64_t       i_qwFreq,
                 IP7_Trace     *i_pP7Trace,
                 IP7_Telemetry *i_pP7Tel,
                 IProxyClient  *i_pClient
                );

    virtual ~CProxyStream();
    void     EnableTimeSync();
    void     DisableTimeSync();
    bool     Maintain();
    bool     SendChunks();
    int32_t  Add_Ref();
    int32_t  Release();

    void     On_Init(sP7C_Channel_Info *i_pInfo);

    void     ProcessTimeSyncResponse(stuP7timeSynResHdr *i_pTime);

    template<typename HdrType>
    void ProcessCreateTimeResponse(stuP7timeResHdr *i_pTime, HdrType i_pP7hdr)
    {
        if (eStateTimeSync == m_eState)
        {
            double l_dbTimeDiff = 0.0;
            if (!m_bConvertEndianess)
            {
                i_pP7hdr->qwTimer_Frequency = i_pTime->qwCpuFreq;
                i_pP7hdr->qwTimer_Value     = i_pTime->qwCpuStartTime;
                l_dbTimeDiff = (double)i_pTime->qwCpuCurrenTime - (double)i_pTime->qwCpuStartTime;

            }
            else
            {
                i_pP7hdr->qwTimer_Frequency = ntohqw(i_pTime->qwCpuFreq);
                i_pP7hdr->qwTimer_Value     = ntohqw(i_pTime->qwCpuStartTime);
                l_dbTimeDiff = (double)ntohqw(i_pTime->qwCpuCurrenTime) - (double)i_pP7hdr->qwTimer_Value;
            }

            m_qwCpuHostTimeOffset = (int64_t)l_dbTimeDiff;

            m_qwCpuStartTime = i_pP7hdr->qwTimer_Value;
            m_qwCpuFreq      = i_pP7hdr->qwTimer_Frequency;

            //calculate diff between creation & current CPU time to set channel creation offset
            l_dbTimeDiff = l_dbTimeDiff * (double)TIME_SEC_100NS / (double)i_pP7hdr->qwTimer_Frequency;
            uint64_t l_qwChannelTime = (((uint64_t)i_pP7hdr->dwTime_Hi) << 32) + (uint64_t)i_pP7hdr->dwTime_Lo;
            l_qwChannelTime = (uint64_t)((double)l_qwChannelTime - l_dbTimeDiff);

            i_pP7hdr->dwTime_Hi = (uint32_t)((l_qwChannelTime >> 32) & 0xFFFFFFFFull);
            i_pP7hdr->dwTime_Lo = (uint32_t)(l_qwChannelTime & 0xFFFFFFFFull);

            m_eState = eStateReady;
        }
    }
};

#endif //UP7_PROXY_STREAM_H